"""A module that provides some Bayesian analysis oriented utility code."""

from __future__ import annotations

from collections.abc import Mapping
from functools import wraps
from io import StringIO
from typing import TYPE_CHECKING, Annotated, Any, NewType, ParamSpec

import pandas as pd
from pydantic import PlainSerializer, PlainValidator
from stanio.json import process_dictionary

if TYPE_CHECKING:
    from collections.abc import Callable

CoordDict = NewType("CoordDict", dict[str, list[str]])
StanInputDict = Mapping[str, Any]

P = ParamSpec("P")


def validate_df_or_string(v: pd.DataFrame | str) -> pd.DataFrame:
    """Load a dataframe even if it is in json string form."""
    if isinstance(v, str):
        v = pd.read_json(StringIO(v))
    return v


# A type for DataFrames in pydantic models that makes them validate and json
# serialize nicely. see https://github.com/pydantic/pydantic/discussions/4243
DfInPydanticModel = Annotated[
    pd.DataFrame,
    PlainValidator(validate_df_or_string),
    PlainSerializer(lambda x: x.to_json(), when_used="always"),
]


def returns_stan_input(
    func: Callable[P, Mapping[str, Any]],
) -> Callable[P, Mapping[str, Any]]:
    """Decorate a function so it returns a json-serialisable dictionary."""

    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> Mapping[str, Any]:
        return process_dictionary(func(*args, **kwargs))

    return wrapper


def one_encode(s: pd.Series) -> pd.Series:
    """Replace a series's values with 1-indexed integer factors.

    :param s: a pandas Series that you want to factorise.

    """
    return pd.Series(pd.factorize(s)[0] + 1, index=s.index)


def make_columns_lower_case(df: pd.DataFrame) -> pd.DataFrame:
    """Make a DataFrame's columns lower case.

    :param df: a pandas DataFrame
    """
    new = df.copy()
    if isinstance(new.columns, pd.MultiIndex):
        new.columns = pd.MultiIndex.from_arrays(
            [
                [c.lower() for c in new.columns.get_level_values(i)]
                for i in range(len(new.columns.levels))
            ],
        )
    else:
        new.columns = pd.Index([c.lower() for c in new.columns])
    return new



"""Functions for running inferences."""

import logging
from collections.abc import Callable
from enum import Enum
from pathlib import Path

import arviz as az

from bibat.fitting_mode import FittingMode
from bibat.inference_configuration import (
    InferenceConfiguration,
    load_inference_configuration,
)
from bibat.prepared_data import PreparedData


class IdataSaveFormat(str, Enum):
    """An enum for choosing the group that a fitting mode writes to."""

    zarr = "prior"
    json = "json"


def run_all_inferences(  # noqa: PLR0913
    inferences_dir: Path,
    data_dir: Path,
    fitting_mode_options: dict[str, FittingMode],
    loader: Callable[[Path], PreparedData],
    local_functions: dict[str, Callable],
    idata_save_format: IdataSaveFormat = IdataSaveFormat.zarr,
) -> None:
    """Fit all inferences in all modes."""
    inference_dirs = inferences_dir.iterdir()
    for inference_dir in sorted(inference_dirs):
        ic = load_inference_configuration(inference_dir)
        prepared_data_json = (data_dir / ic.prepared_data).with_suffix(".json")
        prepared_data = loader(prepared_data_json)
        idata = run_inference(
            ic,
            prepared_data,
            fitting_mode_options,
            local_functions,
        )
        if idata_save_format == IdataSaveFormat.zarr:
            idata_dir = inference_dir / "idata"
            idata.to_zarr(str(idata_dir))
        else:
            idata_file = inference_dir / "idata.json"
            logging.info("Saving idata to %s", idata_file)
            az.to_json(idata, idata_file)


def run_inference(
    ic: InferenceConfiguration,
    prepared_data: PreparedData,
    fitting_mode_options: dict[str, FittingMode],
    local_functions: dict[str, Callable],
) -> az.InferenceData:
    """Run an inference."""
    idata_kwargs = {
        "log_likelihood": "llik",
        "coords": prepared_data.coords,
        "dims": ic.dims,
    }
    if ic.stan_input_function is not None:
        idata_kwargs["observed_data"] = local_functions[ic.stan_input_function](
            prepared_data,
        )
    llik_outputs: dict = {}
    for mode_name in ic.fitting_modes:
        mode = fitting_mode_options[mode_name]
        output = mode.fit(ic, prepared_data, local_functions)
        if mode.idata_target in ["prior", "posterior"]:
            idata_kwargs[mode.idata_target] = output
            idata_kwargs[f"{mode.idata_target.value}_predictive"] = "yrep"
        elif mode.idata_target == "log_likelihood":
            llik_outputs[f"llik_{mode.name}"] = output
    idata = az.from_cmdstanpy(**idata_kwargs)
    for varname, output in llik_outputs.items():
        idata.log_likelihood[varname] = output
    return idata



"""The inference_configuration module.

This module provides the class InferenceConfiguration and the function
`load_inference_configuration`.

"""

from __future__ import annotations

from pathlib import Path

import toml
from pydantic import BaseModel, Field, field_validator, model_validator

DEFAULT_DIMS = {"llik": ["observation"], "yrep": ["observation"]}
DEFAULT_SAMPLE_KWARGS = {"show_progress": False}


class InferenceConfiguration(BaseModel):
    """Configuration for a statistical inference.

    :param name: A name identifying the model configuration

    :param stan_file: Path to a Stan program, with "/" even on windows

    :param prepared_data: Name of the prepared data for this inference

    :param stan_input_function: name of a function from src.stan_input_functions
    used to get a Stan input dictionary from a PreparedData object.

    :param sample_kwargs: dictionary of keyword arguments to
    cmdstanpy.CmdStanModel.sample.

    :param modes: which modes to run the model in. Choose one or more of the
    AVAILABLE_MODES.

    :param dims: map from parameter names to lists of coordinate names.

    :param cpp_options: valid choices for the `cpp_options` argument to
    CmdStanModel

    :param stanc_options: valid choices for the `cpp_options` argument to
    CmdStanModel
    """

    name: str
    stan_file: str
    prepared_data: str
    stan_input_function: str
    fitting_modes: list[str] = Field(alias="modes")
    sample_kwargs: dict = Field(default_factory=lambda: DEFAULT_SAMPLE_KWARGS)
    dims: dict[str, list[str]] = Field(default_factory=lambda: DEFAULT_DIMS)
    mode_options: dict[str, dict] = Field(default_factory=dict)
    cpp_options: dict | None = None
    stanc_options: dict | None = None

    @model_validator(mode="after")
    def check_folds(self: InferenceConfiguration) -> InferenceConfiguration:
        """Check that there is a number of folds if required."""
        if any(m == "kfold" for m in self.fitting_modes):
            if self.mode_options == {}:
                msg = "Mode 'kfold' requires a mode_options.kfold table."
                raise ValueError(
                    msg,
                )
            if "kfold" not in self.mode_options:
                msg = "Mode 'kfold' requires a mode_options.kfold table."
                raise ValueError(
                    msg,
                )
            if "n_folds" not in self.mode_options["kfold"]:
                msg = "Set 'n_folds' field in kfold mode options."
                raise ValueError(msg)
            mo = self.mode_options["kfold"]["n_folds"]
            if isinstance(mo, str) and not mo.isdigit():
                msg = (
                    f"Could not coerce n_folds choice "
                    f"{self.mode_options['kfold']['n_folds']} to int."
                )
                raise ValueError(msg)
        return self

    @field_validator("stan_file")
    @classmethod
    def check_stan_file_exists(
        cls: type[InferenceConfiguration],
        v: str,
    ) -> str:
        """Check that the stan file exists."""
        stan_dir = Path("src") / "stan"
        file = stan_dir / v
        if not file.exists():
            msg = f"{v} is not a file in {stan_dir}."
            raise ValueError(msg)
        return v


def load_inference_configuration(path: Path) -> InferenceConfiguration:
    """Load an inference configuration object from a toml file.

    :param path: Path to directory containing a suitable config.toml file

    """
    kwargs = toml.load(path / "config.toml")
    for k, default in zip(
        ["dims", "sample_kwargs"],
        [DEFAULT_DIMS, DEFAULT_SAMPLE_KWARGS],
        strict=True,
    ):
        if k in kwargs:
            kwargs[k] = default | kwargs[k]
    return InferenceConfiguration(**kwargs)



"""Provides the base class PreparedData."""

from pydantic import BaseModel, ConfigDict

from bibat.util import CoordDict


class PreparedData(BaseModel):
    """What prepared data looks like in a bibat analysis."""

    name: str
    coords: CoordDict
    model_config = ConfigDict(arbitrary_types_allowed=True)



"""A general definition of a fitting mode, plus some mode instances."""

from __future__ import annotations

from collections.abc import Callable  # noqa: TCH003
from enum import Enum
from pathlib import Path

import numpy as np
import xarray as xr
from cmdstanpy import CmdStanMCMC, CmdStanModel
from pydantic import BaseModel
from sklearn.model_selection import KFold

from bibat.inference_configuration import InferenceConfiguration  # noqa: TCH001
from bibat.prepared_data import PreparedData  # noqa: TCH001


class IdataTarget(str, Enum):
    """An enum for choosing the group that a fitting mode writes to."""

    prior = "prior"
    posterior = "posterior"
    log_likelihood = "log_likelihood"


class FittingMode(BaseModel):
    """A way of fitting a statistical model.

    Inferences can be configured to use any of the fitting modes defined here
    by including them by name in the top-level list 'modes'. For example:

    ```toml
      ...
      modes = ['prior', 'posterior', 'kfold']
      ...
    ```

    :param name: A string identifying the fitting mode

    :param idata_target: A string identifying the
    [`InferenceData`](https://python.arviz.org/en/stable/api/inference_data.html)
    group that the mode writes to. Must be one of "prior", "posterior" or
    "log_likelihood".

    :param fit: A function that takes in an `InferenceConfiguration` object, a
    `PreparedData` object and a dictionary of local functions, and returns
    either a CmdStanMCMC object (if the `idata_target` is "prior" or
    "posterior") or an xarray DataArray object (if the `idata_target` is
    "log_likelihood")
    """

    name: str
    idata_target: IdataTarget
    fit: Callable[
        [InferenceConfiguration, PreparedData, dict[str, Callable]],
        CmdStanMCMC | xr.DataArray,
    ]


def sample_hmc_prior(
    ic: InferenceConfiguration,
    data: PreparedData,
    local_functions: dict[str, Callable],
) -> CmdStanMCMC:
    """Run hmc in prior mode."""
    sif = local_functions[ic.stan_input_function]
    input_dict = sif(data) | {"likelihood": 0}
    stan_file = Path("src") / "stan" / ic.stan_file
    model = CmdStanModel(stan_file=stan_file)
    sample_kwargs = ic.sample_kwargs
    if ic.mode_options is not None and "prior" in ic.mode_options:
        sample_kwargs |= ic.mode_options["prior"]
    return model.sample(input_dict, **sample_kwargs)


def sample_hmc_posterior(
    ic: InferenceConfiguration,
    data: PreparedData,
    local_functions: dict[str, Callable],
) -> CmdStanMCMC:
    """Run hmc in posterior mode."""
    sif = local_functions[ic.stan_input_function]
    input_dict = sif(data) | {"likelihood": 1}
    stan_file = Path("src") / "stan" / ic.stan_file
    model = CmdStanModel(stan_file=stan_file)
    sample_kwargs = ic.sample_kwargs
    if ic.mode_options is not None and "posterior" in ic.mode_options:
        sample_kwargs |= ic.mode_options["posterior"]
    return model.sample(input_dict, **sample_kwargs)


def sample_hmc_kfold(
    ic: InferenceConfiguration,
    data: PreparedData,
    local_functions: dict[str, Callable],
) -> xr.DataArray:
    """Do k-fold cross validation, given a CmdStanModel, some data and config.

    :param model: a CmdStanModel. It must have a data variables called
    'likelihood', 'N_train', 'N_test', 'ix_train' and 'ix_test'.

    :param input_dict: a Stan input dictionary. It doesn't need to have any of
    the required variables set (they will be overwritten if they are set).

    :param kwargs: dictionary with an entry for 'n_folds' that specifies the
    value of k for k-fold cross-validation, plus keyword arguments for
    CmdStanModel.sample

    """
    k = ic.mode_options["kfold"]["n_folds"]
    kf = KFold(k, shuffle=True, random_state=1234)
    sif = local_functions[ic.stan_input_function]
    input_dict = sif(data) | {"likelihood": 1}
    stan_file = Path("src") / "stan" / ic.stan_file
    model = CmdStanModel(stan_file=stan_file)
    sample_kwargs = ic.sample_kwargs | {
        k: v for k, v in ic.mode_options["kfold"].items() if k != "n_folds"
    }
    lliks_by_fold = []
    full_ix = np.array(input_dict["ix_train"])
    for fold, (ix_train, ix_test) in enumerate(kf.split(full_ix)):
        input_dict_fold = input_dict | {
            "likelihood": 1,
            "N_train": len(ix_train),
            "N_test": len(ix_test),
            "ix_train": full_ix[ix_train].tolist(),
            "ix_test": full_ix[ix_test].tolist(),
        }
        mcmc = model.sample(data=input_dict_fold, **sample_kwargs)
        llik_fold = mcmc.draws_xr(vars=["llik"])
        # remember the fold
        llik_fold["fold"] = fold
        llik_fold = llik_fold.set_coords("fold")
        # set value of index "chain" to zero to match arviz convention
        llik_fold = llik_fold.assign_coords(
            {"new_chain": ("chain", [0])},
        ).set_index(
            chain="new_chain",
        )
        lliks_by_fold.append(llik_fold["llik"])
    return xr.concat(lliks_by_fold, dim="llik_dim_0").sortby("llik_dim_0")


prior_mode = FittingMode(
    name="prior",
    idata_target=IdataTarget.prior,
    fit=sample_hmc_prior,
)
posterior_mode = FittingMode(
    name="posterior",
    idata_target=IdataTarget.posterior,
    fit=sample_hmc_posterior,
)
kfold_mode = FittingMode(
    name="kfold",
    idata_target=IdataTarget.log_likelihood,
    fit=sample_hmc_kfold,
)
